/*
 * java.lang.UNIXProcess.c
 *
 * Copyright (c) 1996-7 T. J. Wilkinson & Associates, London, UK.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * Written by Tim Wilkinson <tim@tjwassoc.demon.co.uk>
 */

#define	DBG(s)

#include "config.h"
#include "config-std.h"
#include "config-io.h"
#include "config-signal.h"
#include "files.h"
#include "java.io.stubs/FileDescriptor.h"
#include "java.lang.stubs/UNIXProcess.h"

extern void* checked_calloc(size_t, size_t);

void
java_lang_UNIXProcess_exec(struct Hjava_lang_UNIXProcess* this, HArrayOfObject* args, HArrayOfObject* envs)
{
	char** argv;
	char** arge;
	int i;
	char* path;
	int arglen;
	int envlen;
	int pid;
	int in[2];
	int out[2];
	int err[2];

	arglen = (args ? obj_length(args) : 0);
	envlen = (envs ? obj_length(envs) : 0);

DBG(	printf("args %d envs %d\n", arglen, envlen); fflush(stdout);	)

	if (arglen < 1) {
		SignalError(0, "java.io.IOException", "No such file");
	}
	path = makeCString((struct Hjava_lang_String*)unhand(args)->body[0]);
	/* Check program exists and we can execute it */
        i = access(path, X_OK);
	if (i < 0) {
		checked_free(path);
		SignalError(0, "java.io.IOException", SYS_ERROR);
	}

	/* Create the pipes to communicate with the child */
	pipe(in);
	pipe(out);
	pipe(err);

	/* Plumb it all in */
	unhand(this)->sync_fd = 0;
	unhand(this)->stdin_fd = (struct Hjava_io_FileDescriptor*)AllocObject("java/io/FileDescriptor");
	unhand(this)->stdout_fd = (struct Hjava_io_FileDescriptor*)AllocObject("java/io/FileDescriptor");
	unhand(this)->stderr_fd = (struct Hjava_io_FileDescriptor*)AllocObject("java/io/FileDescriptor");
	unhand(unhand(this)->stdin_fd)->fd = in[1];
	unhand(unhand(this)->stdout_fd)->fd = out[0];
	unhand(unhand(this)->stderr_fd)->fd = err[0];
	unhand(this)->stdin_stream = (struct Hjava_io_OutputStream*)execute_java_constructor(0, "java.io.FileOutputStream", 0, "(Ljava/io/FileDescriptor;)V", unhand(this)->stdin_fd);
	unhand(this)->stdout_stream = (struct Hjava_io_InputStream*)execute_java_constructor(0, "java.io.FileInputStream", 0, "(Ljava/io/FileDescriptor;)V", unhand(this)->stdout_fd);
	unhand(this)->stderr_stream = (struct Hjava_io_InputStream*)execute_java_constructor(0, "java.io.FileInputStream", 0, "(Ljava/io/FileDescriptor;)V", unhand(this)->stderr_fd);

	/* Program exists, so we can now work off the child to do the rest */
	switch (java_lang_UNIXProcess_fork(this)) {
	case 0:
		/* Child */
		break;
	case -1:
		/* Error */
		close(in[0]);
		close(in[1]);
		close(out[0]);
		close(out[1]);
		close(err[0]);
		close(err[1]);
		SignalError(0, "java.io.IOException", "Fork failed");
		break;
	default:
		/* Parent */
		close(in[0]);
		close(out[1]);
		close(err[1]);
		return;
	}

	/* Tie in the pipes to stdin/out/err */
	dup2(in[0], 0);
	dup2(out[1], 1);
	dup2(err[1], 2);
	close(in[0]);
	close(in[1]);
	close(out[0]);
	close(out[1]);
	close(err[0]);
	close(err[1]);

	/* Build arguments and environment */
	argv = checked_calloc(arglen + 1, sizeof(char*));
	arge = checked_calloc(envlen + 1, sizeof(char*));
	for (i = 0; i < arglen; i++) {
		argv[i] = makeCString((struct Hjava_lang_String*)unhand(args)->body[i]);
	}
	for (i = 0; i < envlen; i++) {
		arge[i] = makeCString((struct Hjava_lang_String*)unhand(envs)->body[i]);
	}
	/* Execute program */
#if defined(HAVE_EXECVE)
	execve(path, argv, arge);
#else
	abort();
#endif
	exit(-1);
}

jint
java_lang_UNIXProcess_fork(struct Hjava_lang_UNIXProcess* this)
{
	int pid;

#if defined(HAVE_FORK)
	pid = fork();
	switch (pid) {
	case 0:
	case -1:
		break;
	default:
		unhand(this)->isalive = 1;
		unhand(this)->exit_code = 0;
		unhand(this)->pid = pid;
	}
#else
	abort();
#endif
	return (pid);
}

jint
java_lang_UNIXProcess_waitForUNIXProcess(struct Hjava_lang_UNIXProcess* this)
{
	int status;

	if (unhand(this)->isalive == 0) {
#if defined(HAVE_WAIT4)
		if (wait4(unhand(this)->pid, &status, 0, 0) != -1) {
			unhand(this)->exit_code = status;
			unhand(this)->isalive = 1;
			unhand(this)->exit_code = 0;
		}
#else
		abort();
#endif
	}
	return (unhand(this)->exit_code);
}

void
java_lang_UNIXProcess_destroy(struct Hjava_lang_UNIXProcess* this)
{
#if defined(HAVE_KILL)
	kill(unhand(this)->pid, SIGTERM);
#else
	abort();
#endif
}
